/*
 * Copyright (C) 2023 PulkoMandy <pulkomandy@pullkomandy.tk>
 *
 * Distributed under terms of the MIT license.
 */

#include "HyperTextView.h"
#include "IdentityWindow.h"

#include <Application.h>
#include <Box.h>
#include <Button.h>
#include <ByteOrder.h>
#include <Directory.h>
#include <Entry.h>
#include <File.h>
#include <GroupLayout.h>
#include <LayoutBuilder.h>
#include <ListView.h>
#include <OptionPopUp.h>
#include <SeparatorView.h>
#include <String.h>
#include <StringView.h>
#include <TextEncoding.h>
#include <Window.h>

#include <errno.h>
#include <stdio.h>

#include <string>
#include <vector>

#define COLOR "\x1B[33m"
#define NORMAL "\x1B[0m"

void disassemble(const std::vector<uint8_t>& script, BString& output)
{
	size_t pc = 0;
	BString buffer;

	while (pc < script.size()) {
		uint8_t op = script[pc];
		switch(op) {
			case 0x00:
				output += "Return ";
				pc += 1;
				break;
			case 0x01:
				buffer.SetToFormat("IsSet(%02x): ", script[pc + 1]);
				output += buffer;
				pc += 2;
				break;
			case 0x03:
				buffer.SetToFormat("HaveItem(%02x): ", script[pc + 1]);
				output += buffer;
				pc += 2;
				break;
			case 0x09:
				buffer.SetToFormat("IfEqual(WORDS[%02x], %02x%02x) ", script[pc + 1], script[pc + 3], script[pc + 2]);
				output += buffer;
				pc += 4;
				break;
			case 0x0A:
				buffer.SetToFormat("IfLess(WORDS[%02x], %02x%02x) ", script[pc + 1], script[pc + 3], script[pc + 2]);
				output += buffer;
				pc += 4;
				break;
			case 0x0C:
				buffer.SetToFormat("Random(BYTES[%02x], %02x): ", script[pc + 2], script[pc + 1]);
				output += buffer;
				pc += 2;
			case 0x0D:
				buffer.SetToFormat("Hook(%02x%02x): ", script[pc + 2], script[pc + 1]);
				output += buffer;
				pc += 3;
				break;
			case 0x0E:
				buffer.SetToFormat("GlobalHook(%02x): ", script[pc + 1]);
				output += buffer;
				pc += 2;
				break;
			case 0x7E:
				output += "OR ";
				pc += 1;
				break;
			case 0x7F:
				// There is at least one script with Hook immediately followed by HaveItem which
				// seems to mean this is not required to form an AND condition. So it must be a NOT
				// This breaks all scripts that I already got working however, so the meaning of
				// all conditionals must be reversed? Which is strange, because HaveItem made more
				// sense than the opposite for example.
				output += "NOT ";
				pc += 1;
				break;
			case 0x80:
				buffer.SetToFormat("80(%02x) ", script[pc + 1]);
				output += buffer;
				pc += 2;
				break;
			case 0x81:
				buffer.SetToFormat("NEG(BYTES[%02x]) ", script[pc + 1]);
				output += buffer;
				pc += 2;
				break;
			case 0x82:
				buffer.SetToFormat("Clear(BYTES[%02x]) ", script[pc + 1]);
				output += buffer;
				pc += 2;
				break;
			case 0x83:
				buffer.SetToFormat("GetItem(%02x) ", script[pc + 1]);
				output += buffer;
				pc += 2;
				break;
			case 0x85:
				// What is the second parameter?
				buffer.SetToFormat("LoseItem(%02x %02x) ", script[pc + 1], script[pc + 2]);
				output += buffer;
				pc += 3;
				break;
			case 0x87:
				buffer.SetToFormat("Set(WORDS[%02x], %02x%02x) ", script[pc + 1], script[pc + 3], script[pc + 2]);
				output += buffer;
				pc += 4;
				break;
			case 0x89:
				buffer.SetToFormat("Add(WORDS[%02x], %02x%02x) ", script[pc + 1], script[pc + 3], script[pc + 2]);
				output += buffer;
				pc += 4;
				break;
			case 0x8A:
				buffer.SetToFormat("Sub(WORDS[%02x], %02x%02x) ", script[pc + 1], script[pc + 3], script[pc + 2]);
				output += buffer;
				pc += 4;
				break;
			case 0x8F:
				buffer.SetToFormat("GoRoom(%02x) ", script[pc + 1]);
				output += buffer;
				pc += 2;
				break;
			case 0x90:
				buffer.SetToFormat("GoScreen(%02x) ", script[pc + 1]);
				output += buffer;
				pc += 2;
				break;
			case 0x91:
				buffer.SetToFormat("ShowGlobalScreen(%02x) ", script[pc + 1]);
				output += buffer;
				pc += 2;
				break;
			case 0x93:
				// Sets the "home screen" for a room, that will be displayed if no hook sets a
				// specific screen
				buffer.SetToFormat("HomeScreen(%02x) ", script[pc + 1]);
				output += buffer;
				pc += 2;
				break;
			case 0x95:
				buffer.SetToFormat("Set(BYTES[%02x], %02x) ", script[pc + 1], script[pc + 2]);
				output += buffer;
				pc += 3;
				break;
			case 0x97:
				output += "WaitInput ";
				pc++;
				break;
			// I guess 98 is BYTES+= and 99 is BYTES-= ?
			case 0xA0:
				buffer.SetToFormat("Compare(WORDS[%02x], %02x%02x) ", script[pc + 1],
					script[pc + 3], script[pc + 2]);
				output += buffer;
				pc += 4;
				break;
			case 0xFA:
				output += "End.";
				pc++;
				break;
			case 0xFC:
				output += "Game Over ";
				pc++;
				break;
			case 0xFD:
				output += "Next Script ";
				pc++;
				break;
			case 0xFE:
				buffer.SetToFormat("Goto(%02x)", script[pc + 1]);
				pc += 2;
				break;
			case 0xFF:
				output += "You won! ";
				pc++;
				break;
			default:
				buffer.SetToFormat("%02x ", op);
				output += buffer;
				pc++;
				break;
		}
	}
}

const char* cp850_lookup[0x80] = {
	"Ç", "ü", "é", "â", "ä", "à", "å", "ç", "ê", "ë", "è", "ï", "î", "ì", "Ä", "Å", "É", "æ", "Æ",
	"ô", "ö", "ò", "û", "ù", "ÿ", "Ö", "Ü", "ø", "£", "Ø", "×", "ƒ", "á", "í", "ó", "ú", "ñ", "Ñ",
	"ª", "º", "¿", "®", "¬", "½", "¼", "¡", "«", "»", "░", "▒", "▓", "│", "┤", "Á", "Â", "À", "©",
	"╣", "║", "╗", "╝", "¢", "¥", "┐", "└", "┴", "┬", "├", "─", "┼", "ã", "Ã", "╚", "╔", "╩", "╦",
	"╠", "═", "╬", "¤", "ð", "Ð", "Ê", "Ë", "È", "ı", "Í", "Î", "Ï", "┘", "┌", "█", "▄", "¦", "Ì",
	"▀", "Ó", "ß", "Ô", "Ò", "õ", "Õ", "µ", "þ", "Þ", "Ú", "Û", "Ù", "ý", "Ý", "¯", "´", "\u00AD",
	"±", "‗", "¾", "¶", "§", "÷", "¸", "°", "¨", "·", "¹", "³", "²", "■", "\u00A0"
};

const rgb_color EGA[] = {
	{0, 0, 0},
	{0, 0, 196},
	{128, 255, 128},
	{128, 255, 255},
	{255, 128, 128},
	{255, 128, 255},
	{255, 255, 128},
	{255, 255, 255},
};

class Room {
	public:
	std::vector<std::string> fMessages;
	std::vector<std::vector<uint8_t>> fScripts;
};

struct ButtonDescriptor {
	uint16_t x;
	uint16_t y;
	uint16_t color;
	uint16_t reserved[3];
	uint16_t trigger;
	BString label;
	BString shortcut;
};


struct Object {
	uint8_t reserved[8];
	BString name;
	BString description;

	uint8_t roomActions[3];
	uint8_t inventoryActions[3];

	bool inInventory;
};


std::vector<Room> gRooms;
std::vector<uint8_t> gVar8;
std::vector<uint16_t> gVar16;
std::vector<Object> gObjects;

std::vector<ButtonDescriptor> gButtons;

uint8_t gCurrentRoom;
uint8_t gCurrentScreen;
uint8_t gLine;
uint8_t gHomeScreen;
uint8_t gTrigger;
uint8_t gGlobalTrigger;
uint8_t gSavePC;
uint8_t gGameStatus;
bool gGender;
BString gFirstName;
BString gLastName;

uint32_t gLastRunScript;

// TODO global buttons

class DebugWindow: public BWindow
{
	public:
		DebugWindow()
			: BWindow(BRect(40, 40, 750, 800), "Readingame debugger", B_DOCUMENT_WINDOW,
				B_AUTO_UPDATE_SIZE_LIMITS | B_QUIT_ON_WINDOW_CLOSE)
			, fDecoder("cp-850")
		{
			BBox* roomBox = new BBox("room");
			fRoomList = new BOptionPopUp("room popup", "Room", new BMessage('ROOM'));
			roomBox->SetLabel(fRoomList);

			fScreenDump = new BTextView("screen text", B_WILL_DRAW);

			fDisassembly = new BListView("script disassemly");
			fDisassembly->SetFont(be_fixed_font);

			fScreenList = new BOptionPopUp("screens popup", "Screen", new BMessage('SCRN'));

			BGroupView* roomView = new BGroupView(B_VERTICAL);

			BLayoutBuilder::Group<>(roomView)
				.SetInsets(B_USE_DEFAULT_SPACING)
				.AddGroup(B_HORIZONTAL)
					.Add(fScreenList)
					.AddGlue()
				.End()
				.Add(fScreenDump)
				.Add(fDisassembly)
			.End();

			roomBox->AddChild(roomView);

			for (int i = 0; i < gRooms.size(); i++) {
				BString name;
				if (i == 0)
					name = "General";
				else
					name.SetToFormat("%02d", i);
				fRoomList->AddOption(name.String(), i);
			}

			// TODO room messages
			// TODO room scripts
			// TODO current room/screen (make them bold+button to quickly reach them)

			fGlobalTrigger = new BStringView("global trigger", "Global Trigger: ");
			fRoomTrigger = new BStringView("room trigger", "Room Trigger: ");

			fFlagsList = new BListView("flags");
			fByteList = new BListView("flags");
			fWordList = new BListView("flags");

			for (int i = 0; i < gObjects.size(); i++)
				fFlagsList->AddItem(new BStringItem(""));
			for (int i = 0; i < gVar8.size(); i++)
				fByteList->AddItem(new BStringItem(""));
			for (int i = 0; i < gVar16.size(); i++)
				fWordList->AddItem(new BStringItem(""));
			
			BLayoutBuilder::Group<>(this, B_HORIZONTAL)
				.SetInsets(B_USE_WINDOW_SPACING)
				.Add(roomBox)
				.AddGrid(B_VERTICAL)
					.Add(fGlobalTrigger, 0, 0)
					.Add(fRoomTrigger, 1, 0)
					.Add(fWordList, 0, 1)
					.Add(fByteList, 1, 1)
					.Add(fFlagsList, 0, 2, 2)
				.End()
			.End();
		}

		// TODO allow to show different rooms and screens

		void Update()
		{
			Lock();
			fRoomList->SetValue(gCurrentRoom);

			BString label;
			label.SetToFormat("Global Trigger: %d", gGlobalTrigger);
			fGlobalTrigger->SetText(label);
			label.SetToFormat("Room Trigger: %d", gTrigger);
			fRoomTrigger->SetText(label);

			ShowRoom(gCurrentRoom);

			for (int i = 0; i < gObjects.size(); i++)
			{
				BString formatted;
				// TODO show the object names as well
				char buffer[25];
				size_t out = sizeof(buffer);
				BString& name = gObjects[i].name;
				size_t in = name.Length();
				fDecoder.Decode(name.String(), in, buffer, out);
				formatted.SetToFormat("%02x: %s %s", i, gObjects[i].inInventory ? "✓" : " ", buffer);
				BStringItem* item = (BStringItem*)fFlagsList->ItemAt(i);
				if (item->Text() == formatted)
					item->SetEnabled(false);
				else {
					item->SetEnabled(true);
					item->SetText(formatted);
				}
			}
			fFlagsList->Invalidate();
			for (int i = 0; i < gVar8.size(); i++)
			{
				BString formatted;
				formatted.SetToFormat("%02x: %02x", i, gVar8[i]);
				BStringItem* item = (BStringItem*)fByteList->ItemAt(i);
				if (item->Text() == formatted)
					item->SetEnabled(false);
				else {
					item->SetEnabled(true);
					item->SetText(formatted);
				}
			}
			fByteList->Invalidate();
			for (int i = 0; i < gVar16.size(); i++)
			{
				BString formatted;
				formatted.SetToFormat("%02x: %04x", i, gVar16[i]);
				BStringItem* item = (BStringItem*)fWordList->ItemAt(i);
				if (item->Text() == formatted)
					item->SetEnabled(false);
				else {
					item->SetEnabled(true);
					item->SetText(formatted);
				}
			}
			fWordList->Invalidate();

			Unlock();
		}

		void ShowRoom(int32 room)
		{
			int32 screen = 1;
			if (room == gCurrentRoom)
				screen = gCurrentScreen;

			while (fScreenList->CountOptions() > 0)
				fScreenList->RemoveOptionAt(0);
			for (int i = 0; i < gRooms[room].fMessages.size(); i++)
			{
				BString name;
				name.SetToFormat("%02d", i + 1);
				fScreenList->AddOption(name, i);
			}
			fScreenList->SetValue(screen - 1);
			ShowScreen(screen - 1);

			// update script disassembly (and remove from stdout)
			while (fDisassembly->CountItems() > 0)
				fDisassembly->RemoveItem(0);
			for (int i = 0; i < gRooms[room].fScripts.size(); i++)
			{
				BString output;
				disassemble(gRooms[room].fScripts[i], output);
				BStringItem* item = new BStringItem(output);
				if (((room << 16) | i) == gLastRunScript)
					item->SetEnabled(false);
				fDisassembly->AddItem(item);
			}
		}

		void ShowScreen(int32 screen)
		{
			if ((screen < 0) || (gRooms[fRoomList->Value()].fMessages.size() <= screen)) {
				fScreenDump->SetText("<no screens>");
				return;
			}

			char buffer[80*25] = {0};
			size_t out = sizeof(buffer);
			std::string& str = gRooms[fRoomList->Value()].fMessages[screen];
			size_t in = str.length();
			fDecoder.Decode(str.c_str(), in, buffer, out);
			fScreenDump->SetText(buffer);
		}

		void MessageReceived(BMessage* message)
		{
			switch (message->what) {
				case 'ROOM':
				{
					ShowRoom(message->FindInt32("be:value"));
					break;
				}
				case 'SCRN':
				{
					ShowScreen(message->FindInt32("be:value"));
					break;
				}
				default:
					BWindow::MessageReceived(message);
					break;
			}
		}

	private:
		BOptionPopUp* fRoomList;
		BOptionPopUp* fScreenList;
		BTextView* fScreenDump;
		BListView* fDisassembly;

		BStringView* fGlobalTrigger;
		BStringView* fRoomTrigger;

		BListView* fFlagsList;
		BListView* fByteList;
		BListView* fWordList;

		BPrivate::BTextEncoding fDecoder;
};

class GameWindow: public BWindow
{
	public:
		GameWindow()
			: BWindow(BRect(40, 40, 500, 400), "Readingame", B_DOCUMENT_WINDOW,
				B_AUTO_UPDATE_SIZE_LIMITS | B_QUIT_ON_WINDOW_CLOSE)
		{
			fMainText = new HyperTextView("main text", B_WILL_DRAW);

			// TODO making it non-editable while having custom insets breaks the layout somehow?
			fMainText->MakeEditable(false);
			fMainText->MakeSelectable(false);
			fMainText->SetInsets(10, 10, 10, 10);

			fInlineButton = new BButton("button", NULL);

			BGridView* bottomArea = new BGridView();
			fGrid = bottomArea->GridLayout();
			fGrid->SetSpacing(0, 0);

			for (const auto& button: gButtons)
			{
				BString label(button.label);
				label.RemoveAll("$");
				BButton* widget = new BButton(label, new BMessage(button.trigger));
				// widget->SetLowColor(EGA[button.color]);
				// Done in Show instead because it gets overriden
				fGrid->AddView(widget, button.x, button.y, 12, 1);
			}

			BSeparatorView* sv = new BSeparatorView(B_HORIZONTAL);
			sv->SetExplicitMinSize(BSize(515, 16));
			fGrid->AddView(sv, 0, 0, 80);

			for (int i = 0; i < 80; i++)
			{
				fGrid->SetMinColumnWidth(i, 4);
			}

			BLayoutBuilder::Group<>(this, B_VERTICAL)
				.SetInsets(0, 0, 0, B_USE_WINDOW_SPACING)
				.Add(fMainText, 1)
				.AddGroup(B_HORIZONTAL, B_USE_DEFAULT_SPACING, 0)
					.GetLayout(&fLayout)
					.AddGlue()
					.Add(fInlineButton)
					.AddGlue()
				.End()
				.Add(bottomArea)
			.End();
		}

		void Show() override
		{
			BWindow::Show();
			Lock();
			fInlineButton->SetLowColor(make_color(255, 128, 255));
			fMainText->SetViewColor(make_color(196, 255, 255));
			fInlineButton->SetTarget(be_app);

			for (int i = 0; i < gButtons.size(); i++)
			{
				BView* v = ((BLayout*)fGrid)->ItemAt(i)->View();
				v->SetLowColor(EGA[gButtons[i].color]);
				dynamic_cast<BControl*>(v)->SetTarget(be_app);
			}

			Unlock();
		}

		void SetButtonVisible(bool visible)
		{
			int32 index = fLayout->IndexOfView(fInlineButton);
			fLayout->ItemAt(index)->SetVisible(visible);
			//fInlineButton->SetEnabled(visible);
		}

		void SetMainText(const std::string& text)
		{
			Lock();
			SetButtonVisible(false);
			fMainText->SetText("");

			BString decoded;
			PrettyPrint(text, decoded);
			decoded += '\n';

			text_run_array runs;
			runs.count = 1;
			runs.runs->font = be_fixed_font;
			runs.runs->offset = 0;
			runs.runs->color = make_color(0, 0, 0);
			fMainText->Insert(decoded, &runs);
			fMainText->Invalidate();
			Unlock();
		}

		void PrettyPrint(const std::string& message, BString& decoded)
		{
			size_t cursor = 0;
			while (cursor < message.size()) {
				switch(message[cursor]) {
					case '\\':
						cursor += HandleEscape(message, cursor + 1);
						break;
					case '%':
						cursor += HandlePercentEscape(message, cursor + 1, decoded);
						break;
					case '[':
					{
						text_run_array runs;
						runs.count = 1;
						runs.runs->font = be_fixed_font;
						runs.runs->offset = 0;
						runs.runs->color = make_color(0, 0, 0);
						fMainText->Insert(decoded, &runs);
						decoded = "";
						cursor += HandleHyperlink(message, cursor);
						break;
					}
					case '{':
					{
						text_run_array runs;
						runs.count = 1;
						runs.runs->font = be_plain_font;
						runs.runs->offset = 0;
						runs.runs->color = make_color(0, 0, 0);
						fMainText->Insert(decoded, &runs);
						decoded = "";
						cursor += HandleConditional(message, cursor);
						break;
					}
					case '@':
					{
						// @ Gender mark (print an e if the player identifies as female)
						if (gGender)
							decoded += 'e';
						cursor++;
						break;
					}
					case '$':
						// $ Plural mark (print an s if the counter is plural)
						// TODO not implemented yet
						printf(COLOR "%c\n" NORMAL, message[cursor]);
						cursor++;
						break;
					case '_':
						// Used as a non-breakable space
						decoded += "\u00A0";
						cursor++;
						break;
					default:
						if (message[cursor] >= 0)
							decoded += message[cursor];
						else
							decoded += cp850_lookup[(message[cursor] - 0x80) & 0xFF];
						cursor++;
						break;
				}
			}
		}

		int ParseInt(const std::string& message, int& cursor)
		{
			int v = 0;
			while (isdigit(message[cursor])) {
				v = v * 10 + message[cursor] - '0';
				cursor++;
			}

			return v;
		}

		int HandleConditional(const std::string& message, int cursor)
		{
			int icursor = cursor;
			// {%f3|%m2~%m3}
			// {%c3|%m2|%m3|...}
			// {male|female}
			cursor++;
			if (message[cursor] == '%') {
				if ( message[cursor + 1] == 'f') {
					cursor += 2;
					return HandlePercentFConditional(message, cursor) + 3;
				} else {
					fprintf(stderr, "Unknwown message conditional\n");
					return 2;
				}
			}

			int start, end;

			if (gGender) {
				while(message[cursor] != '|')
					cursor++;
				cursor++;
				start = cursor;
				while(message[cursor] != '}')
					cursor++;
				end = cursor;
				cursor++;
			} else {
				start = cursor;
				while(message[cursor] != '|')
					cursor++;
				end = cursor;
				cursor++;
				while(message[cursor] != '}')
					cursor++;
				cursor++;
			}

			std::string extracted = message.substr(start, end - start);
			fMainText->Insert(extracted.c_str());

			return cursor - icursor;
		}

		int HandlePercentFConditional(const std::string& message, int cursor)
		{
			int icursor = cursor;

			int flagToCheck = ParseInt(message, cursor);
			uint8_t valueToCheck = gVar8[flagToCheck];

			if (message[cursor] != '|') {
				fprintf(stderr, "Unknwown message conditional action\n");
				return 1;
			}

			for(;;) {
				if (!valueToCheck && message[cursor] == '|')
					break;
				if (valueToCheck && message[cursor] == '~')
					break;
				cursor++;
			}
			cursor++;

			if (message[cursor] != '%' || message[cursor + 1] != 'm') {
				fprintf(stderr, "Unknwown message output %c%c\n", message[cursor], message[cursor + 1]);
			}
			cursor += 2;

			int messageToDisplay = ParseInt(message, cursor);
			BString decodedMessage;
			PrettyPrint(gRooms[gCurrentRoom].fMessages[messageToDisplay - 1], decodedMessage);
			fMainText->Insert(decodedMessage);

			do {
				cursor++;
			} while (message[cursor] != '}');
			cursor++;

			return cursor - icursor;
		}

		int HandleHyperlink(const std::string& message, int cursor)
		{
			BString linkText;

			int v = 0;
			int i = 1;
			while (isdigit(message[cursor + i])) {
				v = v * 10 + message[cursor + i] - '0';
				i++;
			}

			// TODO links can contain escape sequences (for example @)

			while (message[cursor + i] != ']') {
				linkText += message[cursor + i];
				i++;
			}

			BString decoded;
			PrettyPrint(linkText.String(), decoded);

			text_run_array single_run;
			single_run.count = 1;
			single_run.runs->offset = 0;
			single_run.runs->font = be_plain_font;
			single_run.runs->color = make_color(0, 0, 196);
			fMainText->InsertHyperText(decoded, new HyperTextAction(v), &single_run);

			return i + 1;
		}

		int HandlePercentEscape(const std::string& message, int cursor, BString& decoded)
		{
			if (message[cursor] == 'M') {
				int value = message[cursor + 1] - '0';
				PrettyPrint(gRooms[0].fMessages[value - 1], decoded);
				decoded += '\n';
				return 3;
			} else if (message[cursor] == 'p') {
				decoded += gFirstName;
				return 2;
			} else if (message[cursor] == 'n') {
				decoded += gLastName;
				return 2;
			} else {
				// TODO unknown escape sequence
				fprintf(stderr, "ERROR: unknown percent escape sequence %c\n", message[cursor]);
				return 2;
			}
		}

		int HandleEscape(const std::string& message, int cursor)
		{
			if (message[cursor] == 'o') {
				fInlineButton->SetLabel("OK");
				SetButtonVisible(true);
				int v = 0;
				int i = 1;
				while (isdigit(message[cursor + i])) {
					v = v * 10 + message[cursor + i] - '0';
					i++;
				}

				// If there is no specified value, the default is 0x64
				if (i == 1)
					v = 0x64;

				fInlineButton->SetMessage(new BMessage(v));
				return i + 1;
			} else if (message[cursor] == '!') {
				printf(COLOR "APPEND" NORMAL "\n");
				return 2;
			} else {
				// TODO unknown escape sequence
				fprintf(stderr, "ERROR: unknown escape sequence %c\n", message[cursor]);
				return 2;
			}
		}

	private:
		HyperTextView* fMainText;
		BButton* fInlineButton;
		BGroupLayout* fLayout;
		BGridLayout* fGrid;
};

class Readingame: public BApplication
{
	public:
	Readingame()
		: BApplication("application/x-vnd.PulkoMandy.Readingame")
	{
	}

	void ReadyToRun() override
	{
		BWindow* identity = new IdentityWindow();
		identity->Show();
		identity->CenterOnScreen();

		// TODO check if we have a game loaded
		fMainWindow = new GameWindow();

		// FIXME add the global buttons to the main window

		fDebugWindow = new DebugWindow();


		// TODO: only if asked from the command line
		fDebugWindow->Show();

		// Initial update so we can get the proper highlights for changed variables after running
		// the init
		fDebugWindow->Update();

	}

	void MessageReceived(BMessage* message) override
	{
		if (message->what == 'STRT') {
			fMainWindow->Show();
			fMainWindow->CenterOnScreen();

			bool runNextLine = true;
			int line = 0;
			while (runNextLine && (line < gRooms[0].fScripts.size())) {
				gSavePC = 0;
				runNextLine = RunScriptLine(gRooms[0].fScripts[line]);
				gLastRunScript = line;
				line++;
			}
			fMainWindow->SetMainText(gRooms[gCurrentRoom].fMessages[gCurrentScreen - 1]);
			fDebugWindow->Update();
		} else if (message->what < 256) {
			gTrigger = message->what;

			bool runNextLine = true;

			if (gSavePC != 0) {
				goto resume;
			}

			gLine = 0;
			gCurrentScreen = gHomeScreen;
			while (runNextLine && (gLine < gRooms[0].fScripts.size())) {
				gSavePC = 0;
				runNextLine = RunScriptLine(gRooms[0].fScripts[gLine]);
				gLastRunScript = gLine;
				gLine++;
			}
			gLine = 0;
			while (runNextLine && (gLine < gRooms[gCurrentRoom].fScripts.size())) {
				gSavePC = 0;
resume:
				runNextLine = RunScriptLine(gRooms[gCurrentRoom].fScripts[gLine]);
				gLastRunScript = gLine | (gCurrentRoom << 16);
				if (gSavePC == 0)
					gLine++;
			}

			if (gGameStatus == 0 && gSavePC == 0) {
				fMainWindow->SetMainText(gRooms[gCurrentRoom].fMessages[gCurrentScreen - 1]);
				gLine = 0;
			}

			if (gGameStatus != 0) {
				// TODO go back to main menu, when we have one
				fMainWindow->Lock();
				fMainWindow->SetButtonVisible(false);
				fMainWindow->Unlock();
			}

			fDebugWindow->Update();
		} else {
			BApplication::MessageReceived(message);
		}
	}

	bool RunScriptLine(const std::vector<uint8_t>& script)
	{
		int pc = gSavePC;
		gSavePC = 0;
		bool done = false;
		bool runNextLine = false;
		bool conditionIsInverted = false;

		while (!done) {
			switch(script[pc++]) {
				case 0x00:
				{
					done = true;
					break;
				}
				case 0x01:
				{
					uint8_t varAddress = script[pc++];
					if (conditionIsInverted ^ (gVar8[varAddress] == 0)) {
						done = true;
						runNextLine = true;
					}
					conditionIsInverted = false;
					break;
				}
				case 0x03:
				{
					uint8_t varAddress = script[pc++] - 1;
					if (conditionIsInverted ^ (gObjects[varAddress].inInventory == 0)) {
						done = true;
						runNextLine = true;
					}
					conditionIsInverted = false;
					break;
				}
				case 0x09:
				{
					uint8_t varAddress = script[pc++];
					uint16_t varValue = script[pc++];
					varValue = varValue | (script[pc++] << 8);
					if (conditionIsInverted ^ (gVar16[varAddress] != varValue)) {
						done = true;
						runNextLine = true;
					}
					conditionIsInverted = false;
					break;
				}
				case 0x0C:
				{
					uint8_t maxValue = script[pc++];
					uint8_t varAddress = script[pc++];
					gVar8[varAddress] = random() % maxValue;
					break;
				}
				case 0x0D:
				{
					uint16_t eventToCheck = script[pc++];
					eventToCheck |= script[pc++] << 8;
					if (conditionIsInverted ^ (gTrigger != eventToCheck)) {
						done = true;
						runNextLine = true;
					}
					conditionIsInverted = false;
					break;
				}
				case 0x0E:
				{
					uint8_t eventToCheck = script[pc++];
					// FIXME is it really gTrigger here, or should it be something else?
					if (gGlobalTrigger == eventToCheck) {
						gGlobalTrigger++;
					} else {
						done = true;
						runNextLine = true;
					}
					break;
				}
				case 0x7E:
				{
					// FIXME figure this out
					break;
				}

				case 0x7F:
				{
					conditionIsInverted = true;
					break;
				}
				case 0x81:
				{
					uint8_t varAddress = script[pc++];
					gVar8[varAddress] = !gVar8[varAddress];
					break;
				}
				case 0x82:
				{
					uint8_t varAddress = script[pc++];
					gVar8[varAddress] = 0;
					break;
				}
				case 0x83:
				{
					uint8_t varAddress = script[pc++];
					gObjects[varAddress - 1].inInventory = true;
					break;
				}
				case 0x87:
				{
					uint8_t varAddress = script[pc++];
					uint16_t varValue = script[pc++];
					varValue = varValue | (script[pc++] << 8);
					gVar16[varAddress] = varValue;
					break;
				}
				case 0x8A:
				{
					uint8_t varAddress = script[pc++];
					uint16_t varValue = script[pc++];
					varValue = varValue | (script[pc++] << 8);
					gVar16[varAddress] -= varValue;
					break;
				}
				case 0x8F:
				{
					gCurrentRoom = script[pc++];
					break;
				}
				case 0x90:
				{
					gCurrentScreen = script[pc++];
					break;
				}
				case 0x91:
				{
					fMainWindow->SetMainText(gRooms[0].fMessages[script[pc++]]);
					break;
				}
				case 0x93:
				{
					gHomeScreen = script[pc++];
					gCurrentScreen = gHomeScreen;
					break;
				}
				case 0x95:
				{
					uint8_t varAddress = script[pc++];
					uint8_t varValue = script[pc++];
					gVar8[varAddress] = varValue;
					break;
				}
				case 0x97:
				{
					gSavePC = pc;
					done = true;
					break;
				}
				case 0xA0:
				{
					// FIXME: the global hooks are always evaluated anyways?
					pc += 3;
					break;
				}
				case 0xAE:
				{
					printf("TODO: unhandled opcode AE\n");
					break;
				}
				case 0xFC:
				{
					printf("Game over\n");
					gGameStatus = 1;
					done = true;
					break;
				}
				case 0xFF:
				{
					printf("You won!\n");
					gGameStatus = 2;
					done = true;
					break;
				}
				default:
					fprintf(stderr, "ERROR: unknown opcode %02x at %04x (screen %d)\n", script[pc - 1], pc - 1, gHomeScreen);
					for (int k = 0; k < script.size(); k++) {
						if (k == pc - 1)
							printf("*");
						printf("%02x ", script[k]);
					}
					puts("");
					done = true;
					break;
			}

			if (done && (script[pc] == 0x7E)) {
				done = false;
			}
		}

		return runNextLine;
	}

private:
	GameWindow* fMainWindow;
	DebugWindow* fDebugWindow;
};

void usage()
{
	fprintf(stderr,
		"An engine for playing game originally designed for the Lectures Enjeu application\n\n"
		"Usage: readingame path/to/gamedir/\n"
	);
}

void ReadString(BFile& f, BString& output)
{
	char buffer[256];
	int off = 0;

	do {
		f.Read(buffer + off, 1);
		off++;
	} while (buffer[off - 1] != 0);

	output = buffer;
}

void ReadInit(BFile& f)
{
	uint16_t header[40];
	f.Read(&header, 80);

	for (int i = 0; i < 40; i++)
	{
		header[i] = B_LENDIAN_TO_HOST_INT16(header[i]);
		//printf("%04x ", header[i]);
	}
	//puts("");

	off_t fileSize;
	f.GetSize(&fileSize);
	while (fileSize > 0) {
		uint16_t chunkId;
		f.Read(&chunkId, 2);
		chunkId = B_LENDIAN_TO_HOST_INT16(chunkId);

		fileSize -= 2;

		switch (chunkId) {
			case 1:
			{
				// widgets
				ButtonDescriptor button;

				for(;;) {
					f.Read(&button.x, 2);
					f.Read(&button.y, 2);

					if ((button.x == 0) && (button.y == 0))
						break;

					f.Read(&button.color, 2);
					f.Read(&button.reserved, 6);
					f.Read(&button.trigger, 2);

					char buffer[256];
					int off = 0;

					ReadString(f, button.label);
					ReadString(f, button.shortcut);

					button.x = B_LENDIAN_TO_HOST_INT16(button.x);
					button.y = B_LENDIAN_TO_HOST_INT16(button.y);
					button.trigger = B_LENDIAN_TO_HOST_INT16(button.trigger);

					printf("Button: x %d y %d c %d t %04x %s %s (%04x %04x %04x)\n", button.x, button.y,
						button.color, button.trigger, button.label.String(), button.shortcut.String(),
						button.reserved[0], button.reserved[1], button.reserved[2]);
					gButtons.push_back(button);
				}
				break;
			}
			case 2:
			{
				// actions
				struct Action {
					uint8_t id;
					BString label;
					BString shortcut;
				} action;

				for(;;) {
					f.Read(&action.id, 1);
					if (action.id == 0)
						break;
					ReadString(f, action.label);
					ReadString(f, action.shortcut);
					printf("Action: %02x %s %s\n", action.id, action.shortcut.String(), action.label.String());
				}

				break;
			}
			case 3:
			{
				uint16_t header;
				f.Read(&header, 2);
				header = B_LENDIAN_TO_HOST_INT16(header);
				printf("Object header: %d\n", header);

				Object object;

				for (;;) {
					f.Read(&object.reserved[0], 1);
					if (object.reserved[0] == 0)
						break;
					f.Read(&object.reserved[1], 7);
					ReadString(f, object.name);
					ReadString(f, object.description);
					f.Read(&object.roomActions, 3);
					f.Read(&object.inventoryActions, 3);
					printf("Object: %s (%s) - Room actions %x %x %x / Inventory %x %x %x\n    ",
						object.name.String(), object.description.String(),
						object.roomActions[0], object.roomActions[1], object.roomActions[2],
						object.inventoryActions[0], object.inventoryActions[1], object.inventoryActions[2]);
					for (int j = 0; j < 8; j++)
						printf("%02x ", object.reserved[j]);
					printf("\n");

					gObjects.push_back(object);
				}
				// objects
				break;
			}
			case 4:
			{
				// descriptions
				break;
			}
			case 5:
			{
				// action results
				break;
			}
			case 6:
			{
				// compass
				break;
			}
			case 7:
			{
				// TODO
				break;
			}
			case 8:
			{
				// TODO
				break;
			}
			case 9:
			{
				// TODO
				break;
			}
		}

		// TODO remove when we parse more chunks
		if (chunkId == 3)
			break;
	}
}

void ReadRoomsScripts(BFile& f)
{
	static int roomNumber = 0;

	off_t fileSize;
	f.GetSize(&fileSize);
	while (fileSize > 0) {
		uint16_t roomSize;
		f.Read(&roomSize, 2);
		roomSize = B_LENDIAN_TO_HOST_INT16(roomSize);
		fileSize -= roomSize;

		if (roomSize == 0) {
			fileSize -= 2;
			break;
		}

		if (gRooms.size() < roomNumber + 1)
			gRooms.resize(roomNumber + 1);

		int i = 0;
		while (roomSize > 0) {
			uint16_t scriptSize;

			if (f.Read(&scriptSize, 2) != 2) {
				fprintf(stderr, "ERROR\n");
				return;
			}

			scriptSize = B_LENDIAN_TO_HOST_INT16(scriptSize);
			if (scriptSize == 0) {
				roomSize -= 2;
				break;
			}

			if (gRooms[roomNumber].fScripts.size() < i + 1)
				gRooms[roomNumber].fScripts.resize(i + 1);
			gRooms[roomNumber].fScripts[i].resize(scriptSize - 2);
			f.Read(gRooms[roomNumber].fScripts[i].data(), scriptSize - 2);

			roomSize -= scriptSize;
			i++;
		}
		roomNumber++;
	}
}

void ReadRoomsMessages(BFile& f)
{
	static int roomNumber = 0;

	off_t fileSize;
	f.GetSize(&fileSize);
	while (fileSize > 0) {
		uint16_t roomSize;
		f.Read(&roomSize, 2);
		roomSize = B_LENDIAN_TO_HOST_INT16(roomSize);
		fileSize -= roomSize;

		if (roomSize == 0) {
			fileSize -= 2;
			break;
		}

		roomSize -= 2;

		if (gRooms.size() < roomNumber + 1)
			gRooms.resize(roomNumber + 1);

		int i = 0;
		while (roomSize > 0) {
			uint16_t scriptSize;

			ssize_t r = f.Read(&scriptSize, 2);
			if (r != 2) {
				fprintf(stderr, "ERROR %d (%s)\n", r, strerror(errno));
				return;
			}
			scriptSize = B_LENDIAN_TO_HOST_INT16(scriptSize);
			if (scriptSize == 0) {
				roomSize -= 2;
				break;
			}

			if (gRooms[roomNumber].fMessages.size() < i + 1)
				gRooms[roomNumber].fMessages.resize(i + 1);
			gRooms[roomNumber].fMessages[i].resize(scriptSize - 2);
			f.Read(gRooms[roomNumber].fMessages[i].data(), scriptSize - 2);

			roomSize -= scriptSize;
			i++;
		}
		roomNumber++;
	}
}

int main(int argc, char** argv)
{
	if (argc != 2) {
		// TODO open a filepanel instead to load a game
		// TODO or just look for the TITRES.DAT file
		usage();
		return 1;
	}

	BDirectory gameDir(argv[1]);

	if (gameDir.InitCheck() != B_OK) {
		fprintf(stderr, "Game directory %s not found.\n", argv[1]);
		usage();
		return 1;
	}

	BEntry entry;
	BFile initFile;
	BFile msgRoomsFile;
	BFile msgGeneralFile;
	BFile condRoomsFile;
	BFile condGenFile;
	BFile condUltiFile;

	if (gameDir.FindEntry("INIT.DAT", &entry) == B_OK) {
		initFile.SetTo(&entry, B_READ_ONLY);
	} else {
		fprintf(stderr, "Game does not contain an INIT.DAT file.\n");
		return 2;
	}

	if (gameDir.FindEntry("MSG.DAT", &entry) == B_OK) {
		msgRoomsFile.SetTo(&entry, B_READ_ONLY);
	} else {
		fprintf(stderr, "Game does not contain an MSG.DAT file.\n");
		return 2;
	}

	if (gameDir.FindEntry("MSGGEN.DAT", &entry) == B_OK) {
		msgGeneralFile.SetTo(&entry, B_READ_ONLY);
	} else {
		fprintf(stderr, "Game does not contain an MSGGEN.DAT file.\n");
		return 2;
	}

	if (gameDir.FindEntry("CONDSAL.DAT", &entry) == B_OK) {
		condRoomsFile.SetTo(&entry, B_READ_ONLY);
	} else {
		fprintf(stderr, "Game does not contain a CONDSAL.DAT file.\n");
		return 2;
	}

	if (gameDir.FindEntry("CONDGALE.DAT", &entry) == B_OK) {
		condGenFile.SetTo(&entry, B_READ_ONLY);
	} else {
		fprintf(stderr, "Game does not contain a CONDGALE.DAT file.\n");
		return 2;
	}

	if (gameDir.FindEntry("CONDULTI.DAT", &entry) == B_OK) {
		condUltiFile.SetTo(&entry, B_READ_ONLY);
	} else {
		fprintf(stderr, "Game does not contain a CONDULTI.DAT file.\n");
		return 2;
	}

	ReadInit(initFile);
	ReadRoomsScripts(condGenFile);
	ReadRoomsScripts(condRoomsFile);
	ReadRoomsScripts(condUltiFile);

	ReadRoomsMessages(msgGeneralFile);
	ReadRoomsMessages(msgRoomsFile);

	// TODO how many variables are there?
	gVar8.resize(16);
	gVar16.resize(16);

	Readingame app;
	app.Run();
}
